class ConfigurationException(Exception):
    """
    Exception raised for invalid configurations.

    This exception is used to indicate errors related to configuration validation and processing.
    """
