from .sequence_solvers import SequenceClassificationSolver, SequenceRegressionSolver


# Identical to the SequenceClassificationSolver at the moment
class ResiduesClassificationSolver(SequenceClassificationSolver):
    pass

# Identical to the SequenceRegressionSolver at the moment
class ResiduesRegressionSolver(SequenceRegressionSolver):
    pass